-- This module is used to read out LRB's input and monitoring data.
-- It is to be downloaded into Altera pci1 interface of a DCC daughterboard.
-- (pci1 and pci2 has different DCC side pin assignment, the header format
-- is also different to distinguish between two PCI buses)
-- BAR0 occupies 4 32-bit words of mem space. 
-- 0x0 is used to save the Base Address of MB1's LRB, only bits 31-15 are implemented 
-- 0x4 is used to save the Base Address of MB2's LRB, only bits 31-15 are implemented
-- 0x8 is used to save the Base Address of MB3's LRB, only bits 31-15 are implemented
-- 0xc is LRB's enable register. bits 2-0 enables MB1's channel 2-0
--  							 bits 5-3 enables MB2's channel 2-0
--  							 bits 8-6 enables MB3's channel 2-0
-- LRB monitor data readout takes precedence over LRB input data readout
-- It is started by a mon_latch signal from DCC
-- All enabled LRB are read out. After that, the monitor data
-- are serially output to DCC 
-- dcc_ctrl_o(3) = rqst_mask shift in
-- dcc_ctrl_o(2) = '1' monitor latch signal
-- dcc_ctrl_o(1) = '1' rqst_mask strobe
-- dcc_ctrl_o(0) = '1' reset input from DCC
-- dcc_ctrl_i(3) data word is chnl_en
-- dcc_ctrl_i(2) valid low word flag
-- dcc_ctrl_i(1) monitor data flag
-- dcc_ctrl_i(0) = trailer flag
-- high word always follows low word !!! 
-- 9/23/2005 LRB reset function added. When DCC reset is received, pci interface will initiate a configuration
-- space write with l_adi = x"1c0080"
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity pci_if is
    port (
-- pci clock and reset signals
        clk: in STD_LOGIC;
        rstn: in STD_LOGIC;
-- pci function local side signals
        stat_reg: in STD_LOGIC_VECTOR (5 downto 0);
        l_adro: in STD_LOGIC_VECTOR (31 downto 0);
        l_dato: in STD_LOGIC_VECTOR (31 downto 0);
        l_cmdo: in STD_LOGIC_VECTOR (3 downto 0);
        l_cbeni: out STD_LOGIC_VECTOR (3 downto 0);
        l_adi: out STD_LOGIC_VECTOR (31 downto 0);
-- pci function master local side signal
        lm_tsr: in STD_LOGIC_VECTOR (9 downto 0);
        lm_dxfrn: in STD_LOGIC;
        lm_adr_ackn: in STD_LOGIC;
        lm_ackn: in STD_LOGIC;
        lm_lastn: out STD_LOGIC;
        lm_rdyn: out STD_LOGIC;
        lm_req32n: out STD_LOGIC;
-- pci function target local side signal
        lt_framen: in STD_LOGIC;
        lt_dxfrn: in STD_LOGIC;
        lt_ackn: in STD_LOGIC;
        lt_tsr: in STD_LOGIC_VECTOR (11 downto 0);
        lt_discn: out STD_LOGIC;
        lt_rdyn: out STD_LOGIC;
-- data from/to stc
        dcc_clk: in STD_LOGIC;
        dcc_ctrl_o: in STD_LOGIC_VECTOR (3 downto 0);
        dcc_ctrl_i: out STD_LOGIC_VECTOR (3 downto 0);
        dcc_data: out STD_LOGIC_VECTOR (15 downto 0);
-- link receiver data available
        mb_a: in STD_LOGIC_VECTOR(4 downto 1);
        mb_b: in STD_LOGIC_VECTOR(4 downto 1);
        mb_c: in STD_LOGIC_VECTOR(4 downto 1)
    );
end pci_if;

architecture pci_arch of pci_if is
COMPONENT dcfifo
	GENERIC (LPM_WIDTH: POSITIVE :=35;
		LPM_WIDTHU: POSITIVE :=0;
		LPM_NUMWORDS: POSITIVE := 16;
		LPM_SHOWAHEAD: STRING:= "OFF";
		OVERFLOW_CHECKING: STRING:= "OFF";
		UNDERFLOW_CHECKING: STRING:= "OFF";
		DELAY_RDUSEDW: POSITIVE:= 1;
		DELAY_WRUSEDW: POSITIVE:= 1;
		RDSYNC_DELAYPIPE: POSITIVE:= 3;
		WRSYNC_DELAYPIPE: POSITIVE:= 3);
	PORT (
		data: IN STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0);
		rdclk, wrclk, wrreq, rdreq, aclr: IN STD_LOGIC;
		rdfull,wrfull, wrempty, rdempty: OUT STD_LOGIC;
		q: OUT STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0);
		rdusedw, wrusedw: OUT STD_LOGIC_VECTOR(LPM_WIDTHU-1 DOWNTO 0)
		);
END COMPONENT;
COMPONENT lpm_ram_dp
   GENERIC ( LPM_WIDTH: POSITIVE :=32;
             LPM_WIDTHAD: POSITIVE := 8;
             LPM_NUMWORDS: NATURAL := 256;
             LPM_TYPE: STRING := "LPM_RAM_DP";
             LPM_INDATA: STRING := "REGISTERED";
             LPM_OUTDATA: STRING := "UNREGISTERED";
             LPM_RDADDRESS_CONTROL: STRING := "REGISTERED";
             LPM_WRADDRESS_CONTROL: STRING := "REGISTERED";
             LPM_FILE: STRING := "UNUSED";
             LPM_HINT: STRING := "UNUSED"

           );
   PORT    ( rdaddress, wraddress: IN STD_LOGIC_VECTOR(LPM_WIDTHAD-1 DOWNTO 0);
             rdclock, wrclock: IN STD_LOGIC := '0';
             rden, rdclken, wrclken: IN STD_LOGIC := '1';
             wren: IN STD_LOGIC; 
             data: IN STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0);
             q: OUT STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0));
END COMPONENT;
signal reset: std_logic;
signal reset_LRB: std_logic;
signal reset_LRB_dl: std_logic_vector(2 downto 0);
signal end_reset: std_logic;
signal sel_low: std_logic;
signal wr_fifo: std_logic;
signal rd_fifo: std_logic;
signal fifo_empty: std_logic;
signal fifo_in: std_logic_vector(34 downto 0);
signal fifo_out: std_logic_vector(34 downto 0);
signal header: std_logic;
signal wr_lrb_en: std_logic;
signal wc: std_logic_vector(9 downto 0);
signal tmp_data: std_logic_vector(31 downto 0);
signal clr_wa: std_logic;
signal ec_buf_wa: std_logic;
signal ld_buf_ra: std_logic;
signal ec_buf_ra: std_logic;
signal last_word: std_logic_vector(31 downto 0);
signal sel_tsr0: std_logic;
signal lm_adr_ackn_q: std_logic;
-- priority encoder signals
signal rqst_mask_in: std_logic;
signal rqst_mask_str: std_logic;
signal sr_mask: std_logic_vector(8 downto 0);
signal rqst_mask: std_logic_vector(8 downto 0);
signal rqst_sync: std_logic_vector(8 downto 0);
signal rqst_sync2: std_logic_vector(8 downto 0);
signal scan_cntr: std_logic_vector(3 downto 0);
signal rqst: std_logic;
signal rqst_q: std_logic;
---------------------------
signal chnl_en: std_logic_vector(8 downto 0);
signal chnl_en_rdbk: std_logic_vector(8 downto 0);
signal chnl_addr: std_logic_vector(3 downto 0);
signal chnl_addr_l: std_logic_vector(3 downto 0);
signal lrb_we : std_logic;
signal lrb_wa: std_logic_vector(1 downto 0);
signal lrb_ra: std_logic_vector(1 downto 0);
signal lrb_addr: std_logic_vector(16 downto 0);
signal read_data: std_logic;
signal idle: std_logic;
signal en_master: std_logic;
signal w_cycle: std_logic;
-- lrb monitor data signals
signal en_wr_fifo: std_logic;
signal lrb_en: std_logic_vector(2 downto 0);
signal rd_monitor,mon_latch,mon_active: std_logic;
signal mon_rqst,lm_last,mon_flag: std_logic;
signal lm_last_dl: std_logic;
signal mfifo_wa: std_logic_vector(5 downto 0);
signal mon_active_sync: std_logic_vector(1 downto 0);
signal rd_monitor_sync: std_logic_vector(1 downto 0);
-- dcc clock monitoring
signal dcc_cnt: std_logic_vector(14 downto 0);
signal pci_cnt: std_logic_vector(14 downto 0);
signal dcc_sync: std_logic_vector(3 downto 0);
signal pci_cnt_q: std_logic_vector(14 downto 0);
begin
reset <= dcc_ctrl_o(0);
process(dcc_clk)
begin
  if(dcc_clk'event and dcc_clk = '1')then
    if(sel_low = '0')then
      dcc_data <= fifo_out(31 downto 16);
    else
      dcc_data <= fifo_out(15 downto 0);
    end if;
	 rqst_mask_in <= dcc_ctrl_o(3);
	 rqst_mask_str <= dcc_ctrl_o(1);
  end if;
end process;
process(dcc_clk,reset_LRB)
begin
  if(reset_LRB = '1')then
    rd_fifo <= '0';
    sel_low <= '0';
    mon_latch <= '0';
    dcc_ctrl_i(3 downto 0) <= "0000";
	sr_mask <= (others => '0');
	rqst_mask <= (others => '0');
  elsif(dcc_clk'event and dcc_clk = '1')then
    rd_fifo <= not rd_fifo and not fifo_empty;
    sel_low <= rd_fifo;
    mon_latch <= dcc_ctrl_o(2);
-- load chnl_en word flag
    dcc_ctrl_i(3) <= fifo_out(34) and sel_low;
-- valid low word flag
    dcc_ctrl_i(2) <= not fifo_out(34) and sel_low;
-- monitor data flag
    dcc_ctrl_i(1) <= fifo_out(33);
-- trailer flag
    dcc_ctrl_i(0) <= not fifo_out(32);
	sr_mask <= sr_mask(7 downto 0) & rqst_mask_in;
	if(rqst_mask_str = '1')then
	  rqst_mask <= sr_mask;
	end if;
  end if;
end process;
i_fifo: dcfifo
	   port map(
			data => fifo_in,
			rdclk => dcc_clk,
			wrclk => clk,
			wrreq => wr_fifo,
			rdreq => rd_fifo,
			aclr => reset_LRB,
			rdempty => fifo_empty,
			q => fifo_out);
fifo_in <= wr_lrb_en & mon_flag & lm_tsr(3) & tmp_data;
process(clk)
begin
  if(clk'event and clk = '1')then
    lm_adr_ackn_q <= lm_adr_ackn;
    if(lm_adr_ackn = '0' and lm_adr_ackn_q = '1')then
      mon_flag <= rd_monitor;
      w_cycle <= reset_LRB_dl(1);
      chnl_addr_l <= chnl_addr;
    end if;
    if(lm_adr_ackn = '0')then
      en_wr_fifo <= '0';
    elsif(lm_dxfrn = '0' and w_cycle = '0')then
      en_wr_fifo <= not(mfifo_wa(3) and (mfifo_wa(2) or mfifo_wa(1)));
    end if;
  end if;
end process;
-- write fifo is always delayed one word,i.e.,write the first word when next word comes
-- trailer is written when burst read ends
wr_fifo <= (en_wr_fifo and (not lm_dxfrn or (read_data and not lm_tsr(3)))) or wr_lrb_en;
process(clk)
begin
  if(clk'event and clk = '1')then
    header <= (header and lm_dxfrn and (lm_tsr(2) or lm_tsr(3))) or not lm_adr_ackn;
    if(lm_dxfrn = '0' or lt_dxfrn = '0')then
      if(header = '0' or rd_monitor = '1' or lt_dxfrn = '0')then
        tmp_data <= l_dato;
      else
        tmp_data <= l_dato(31 downto 13) & '0' & chnl_addr_l & l_dato(7 downto 0);
      end if;
    end if;
    if(wr_lrb_en = '1')then
-- dccv5 does not use mip1
      chnl_en <= tmp_data(8 downto 3) & "000";
    end if;
  end if;
end process;
-- LRB channel enable register, controls three LRBs
process(clk,rstn)
begin
  if(rstn = '0')then
    wc <= (others => '0');
    wr_lrb_en <= '0';
  elsif(clk'event and clk = '1')then
    if(lm_adr_ackn = '0')then
      wc <= (others => '0');
    elsif(lm_dxfrn = '0')then
      wc <= wc + 1;
    end if;
    wr_lrb_en <= sel_tsr0 and l_adro(3) and l_adro(2) and l_cmdo(0) and not lt_dxfrn;
  end if;
end process;
-- slave access flags
process(clk)
begin
  if(clk'event and clk = '1')then
    sel_tsr0 <= lt_tsr(0);
  end if;
end process;
-- always insert one wait cycle in slave access, 
lt_rdyn <= not sel_tsr0;
-- tsr0 access is single word only
lt_discn <= lt_dxfrn;
-- mb_x signals are data available signals from LRBs, when header of the last event in a LRB
-- is read out, this signal goes low. It goes high when LRB has at least one complete event
-- only enabled channels are checked
process(clk)
begin
  if(clk'event and clk = '1')then
    rqst_sync <= rqst_mask and (mb_c(3 downto 1) & mb_b(3 downto 1) & mb_a(3 downto 1));
    rqst_sync2 <= chnl_en and rqst_sync;
  end if;
end process;
process(clk,rstn,reset_LRB)
begin
  if (rstn = '0' or reset_LRB = '1')then
	 scan_cntr <= "0000";
	 chnl_addr <= "0000";
	 rqst <= '0';
     rqst_q <= '0';
  elsif(clk'event and clk = '1')then
    if(rqst = '0')then
      scan_cntr(3) <= (scan_cntr(2) and scan_cntr(1)) or (scan_cntr(3) and not scan_cntr(1));
      scan_cntr(2) <= scan_cntr(2) xor (not scan_cntr(3) and scan_cntr(1));
      scan_cntr(1) <= scan_cntr(0);
      scan_cntr(0) <= not scan_cntr(0) and not scan_cntr(1);
	  chnl_addr <= scan_cntr(3) & (not scan_cntr(3) and scan_cntr(2)) & scan_cntr(1) & (not scan_cntr(1) and scan_cntr(0));
	 end if;
--	 if(header = '1' and (lm_dxfrn = '0' or lm_tsr(3 downto 2) = "00"))then
	 if(lm_tsr(2) = '1')then
	   rqst <= '0';
	 elsif(rqst = '0')then
	   case scan_cntr is
		  when "0100" => rqst <= rqst_sync2(3);
		  when "0101" => rqst <= rqst_sync2(4);
		  when "0110" => rqst <= rqst_sync2(5);
		  when "1000" => rqst <= rqst_sync2(6);
		  when "1001" => rqst <= rqst_sync2(7);
		  when "1010" => rqst <= rqst_sync2(8);
		  when others => rqst <= '0';
		end case;
	 end if;
     rqst_q <= rqst and not lm_tsr(2);
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    read_data <= lm_tsr(3);
  end if;
end process;
process(clk,reset)
begin
  if(reset = '1')then
    reset_LRB <= '1';
    reset_LRB_dl <= "000";
  elsif(clk'event and clk = '1')then
    reset_LRB <= reset_LRB and (lm_adr_ackn or not reset_LRB_dl(2));
    reset_LRB_dl <= reset_LRB_dl(1 downto 0) & reset_LRB;
  end if;
end process;
i_lrb_addr: lpm_ram_dp
   generic map (LPM_WIDTH => 17,LPM_WIDTHAD => 2, LPM_NUMWORDS => 4, LPM_FILE => "lrb_addr.mif")
   port map(
           rdaddress => lrb_ra,
           wraddress => l_adro(3 downto 2),
           wrclock => clk,
           rdclock => clk,
           wren => lrb_we,
           data => l_dato(31 downto 15),
           q => lrb_addr);
lrb_ra <= "11" when reset_LRB_dl(1) = '1' else
          mfifo_wa(5 downto 4) when rd_monitor = '1' else
          chnl_addr(3 downto 2) when rqst = '1' else
          l_adro(3 downto 2);
lrb_we <= sel_tsr0 and not(l_adro(3) and l_adro(2)) and not lt_dxfrn and l_cmdo(0);
idle <= not(lm_tsr(3) or lm_tsr(2) or lm_tsr(1) or lm_tsr(0));
-- master logic follows
-- master is always ready when in the data phase 
lm_rdyn <= not (lm_tsr(2) or lm_tsr(3));
lm_lastn <= not lm_last and (lm_dxfrn or not wc(9));
-- request bus when idle and line receiver has data
--lm_req32n <= not (idle and ((((rqst_q and not header) or rd_monitor) and en_master) or reset_LRB_dl(1)));
lm_req32n <= not (idle and (((rqst_q or rd_monitor) and en_master) or reset_LRB_dl(1)));
-- read multiple command and all bytes enabled
l_cbeni <= "1111" when lm_adr_ackn = '1' else "1011" when reset_LRB_dl(1) = '1' else "0110";
process(clk)
begin
  if(clk'event and clk = '1')then
    if(l_adro(3 downto 2) = "11")then
      chnl_en_rdbk <= chnl_en;
    else
      chnl_en_rdbk <= "000000000";
    end if;
  end if;
end process;
l_adi(31 downto 15) <= lrb_addr when l_adro(3) /= l_adro(2) or lm_adr_ackn = '0'
--                       else "00000000000000000";
                       else rqst_mask & mb_c(3 downto 1) & mb_b(3 downto 1) & "00";
-- LRB base address and channle enable register read back
l_adi(14 downto 0) <= "000000" & chnl_en_rdbk when lm_adr_ackn = '1' and l_adro(3 downto 2) /= "00" else
					   pci_cnt_q when lm_adr_ackn = '1' else	
--l_adi(14 downto 0) <= "00000" & chnl_en_rdbk when lm_adr_ackn = '1' else
         "000000010000000" when reset_LRB_dl(1) = '1' else
-- LRB monitor data starting address
         "000000001001000" when rd_monitor = '1' else
-- LRB data fifo starting address 
         '1' & chnl_addr(1 downto 0) & "000000000000";
-- LRB monitor data logic follows
process(clk,reset_LRB)
begin
  if (reset_LRB = '1')then
    mon_active_sync <= "00";
    mon_rqst <= '0';
    rd_monitor <= '0';
    lm_last_dl <= '0';
  elsif(clk'event and clk = '1')then
    mon_active_sync <= mon_active_sync(0) & mon_active;
    mon_rqst <= (mon_rqst and not rd_monitor) or (not mon_active_sync(1) and mon_active_sync(0));
    rd_monitor <= (rd_monitor and not (lm_last_dl and mfifo_wa(5) and mfifo_wa(4))) or (mon_rqst and idle);
    lm_last_dl <= lm_last;
  end if;
end process;
process(dcc_clk,reset_LRB)
begin
  if(reset_LRB = '1')then
    rd_monitor_sync <= "00";
    mon_active <= '0';
  elsif(dcc_clk'event and dcc_clk = '1')then
    rd_monitor_sync <= rd_monitor_sync(0) & rd_monitor;
    mon_active <= en_master and (mon_latch or (mon_active and not(rd_monitor_sync(1) and not rd_monitor_sync(0))));
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    lrb_en(2) <= chnl_en(8) or chnl_en(7) or chnl_en(6);
    lrb_en(1) <= chnl_en(5) or chnl_en(4) or chnl_en(3);
    lrb_en(0) <= chnl_en(2) or chnl_en(1) or chnl_en(0);
    en_master <= lrb_en(2) or lrb_en(1) or lrb_en(0);
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    lm_last <= not lm_dxfrn and mfifo_wa(3) and not mfifo_wa(1) and mfifo_wa(0);
  end if;
end process;
process(clk,rd_monitor)
variable sel_lrb: std_logic_vector(3 downto 0);
begin
  sel_lrb := mfifo_wa(5 downto 4) & lrb_en(2 downto 1);
  if(rd_monitor = '0')then
    mfifo_wa(5 downto 4) <= "11";
  elsif(clk'event and clk = '1')then
    if(mfifo_wa(5 downto 4) = "11")then
      mfifo_wa(5) <= not lrb_en(0) and not lrb_en(1); 
      mfifo_wa(4) <= not lrb_en(0) and (not lrb_en(2) or lrb_en(1)); 
    elsif(lm_last = '1')then 
      case sel_lrb is
        when "0001" | "0011" =>
          mfifo_wa(5 downto 4) <= "01";
        when "0010" | "0111" =>
          mfifo_wa(5 downto 4) <= "10";
        when others =>
          mfifo_wa(5 downto 4) <= "11";
      end case;
    end if; 
  end if;
end process;
process(clk,lm_adr_ackn)
begin
  if(lm_adr_ackn = '0')then
    mfifo_wa(3 downto 0) <= "0000";
  elsif(clk'event and clk = '1')then
    if(lm_dxfrn = '0' and rd_monitor = '1')then
      mfifo_wa(3) <= mfifo_wa(3) xor (mfifo_wa(2) and mfifo_wa(1) and mfifo_wa(0));
      mfifo_wa(2) <= mfifo_wa(2) xor (mfifo_wa(1) and mfifo_wa(0));
      mfifo_wa(1) <= mfifo_wa(1) xor mfifo_wa(0);
      mfifo_wa(0) <= not mfifo_wa(0);
    end if;
  end if;
end process;
-- dcc clock monitoring
process(dcc_clk)
begin
  if(dcc_clk'event and dcc_clk = '1')then
    dcc_cnt <= dcc_cnt + 1;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    dcc_sync <= dcc_sync(2 downto 0) & dcc_cnt(14);
    if(dcc_sync(3 downto 2) = "10")then
      pci_cnt <= "000000000000000";
      pci_cnt_q <= pci_cnt;
    else
      pci_cnt <= pci_cnt + 1;
    end if;
  end if;
end process;
end pci_arch;
